/* This file is a template of the code which can use the API. 
   You can copy and extend the file as you like.
*/

// First, 'usual' include files are specified, e.g.
#ifdef _WINDOWS
#include "stdafx.h"
#endif
#include <stdio.h>
#include <stdlib.h>

/* The symbol 'uniontest' below is used by syntapi.h and propapi.h to specify 
   the action when a dynamic type checking error is detected. It is good practice 
   to detect such errors, at least in a DEBUG version.
*/

#ifdef _DEBUG
#define uniontest(t) fatalerror("Ups ... illegal union field selection")
// You can modify as appropriate
#else
#define uniontest(t) 
#endif

#include <string.h>
#include <syntapi.h> // uses uniontest
#include <propapi.h> // uses uniontest

using namespace CSDL_IR;

/* The fatalerror procedure below is called by the API, when an unforseen situation occurs.
   This procedure must always be defined. After issuing a message, 
   the procedure should throw an exception to be catch somewhere.
*/

void fatalerror(char *str1,char *str2)
{ 
#ifdef _WINDOWS
  CString str;
  str.Format(str1,str2);
  if (AfxMessageBox(str + "\nAbort?",MB_YESNO) == IDYES)
    throw 0;
#else
  fprintf(stdout,str1,str2);
  throw 0;
#endif
}


/* The 'Main' procedure contains the user defined code which
   uses the API. The root class 'Capi' is defined in syntapi.h. 
*/

extern "C" __declspec(dllexport) void Main(Capi *tree);
extern "C" __declspec(dllexport) void GetName(char *str);

int entities[(int)REMOTESCOPE];
int includelinkedpackages = 0;
Capi *thetree = NULL;

void GetName(char *str)
{ strcpy(str,"Detect entities which are un-used");
}

static void setentities()
{ for (int i = 0;i<(int)REMOTESCOPE;i++)
    entities[i] = 1;
}


static seterror(int index,char *text,char **array)
{ if (array[index] == NULL)
    fatalerror("Not enough array indices to include new error message","");
  if (strlen(array[index]) == 0) // test if the message already has been put there
  array[index] = text;
}

static void seterrors(char **errormessages)
{ seterror(0,"Name %s is not used",errormessages);
  seterror(1,"Specification is not fully analyzed",errormessages);
}

static int inpredefinedpackage(Qual *entity)
{ Qual *predefqual = thetree->propertiestreeref->predefquals.predefinedqual;
  while (entity->kind() != REMOTESCOPE)
  { if (entity == predefqual)
      return 1;
    entity = entity->sur();
  }
  return 0;
}

int checkid(Rootclass *r)
{ if (r->treekind() == IDENTIFIER)
  { Id *id = (Id *)r;
    Qual *descriptor = id->descriptor();
    if (descriptor == NULL && id->name() != thetree->syntaxtreeref->errorname && 
        id->name()->position()->symbol()->kind() != QUALIFIER &&
        !inpredefinedpackage(id->name()->position()->symbol()->ownerdiagram()->scopequal()))
    { Sourceposition *pos = id->name()->position();
      Error::usererror(6001,pos->startpos(),pos->endpos(),0,pos->symbol());
      return 0;
    }
    while (descriptor != NULL)
    { descriptor->anamark() = 0;
      descriptor = descriptor->copy();
    }
  }
  return 1;
}


static void setunused(Qual *entitydict)
{ Qual *temp = entitydict;
  while (temp != NULL)
  { temp->anamark() = temp->copy() != NULL || temp->sur()->kind() == REMOTESCOPE?0:1;
    temp = temp->next();
  }
}

static int inlinkedpackage(Qual *entitydict)
{ Symbol *defsym = entitydict->name()->position()->symbol();
  if (defsym->kind() == TEXTSYMBOL)
  { Symbol *textsymbol = defsym;
    if (textsymbol->alt_textsymbolsymbol()->external() != NULL)
      return 1;
  }
  Diagram *d = defsym->ownerdiagram()->firstpage();
  while (d->kind() != REMOTESCOPE)
  { Symbol *refsymbol = d->reference();
    if (refsymbol->prposition() == NULL) // A symbol which not is contained in a text symbol
    { if (refsymbol->alt_referencesymbol() != NULL && refsymbol->alt_referencesymbol() != NULL && refsymbol->alt_referencesymbol()->external() != NULL)
        return 1;
    }
    d = d->ownerdiagram()->firstpage();
  }
  return 0;
}


static void reportunused(Qual *entitydict)
{ while (entitydict != NULL)
  { if (entitydict->anamark() && !entitydict->iserror() && !entitydict->isimplicitqual())
    { Name *entityname = entitydict->name();
      Sourceposition *pos = entityname->position();
      if (includelinkedpackages || !inlinkedpackage(entitydict))
        Error::usererror(6000,pos->startpos(),pos->endpos(),(long)entitydict->spell(),pos->symbol());
    }
    entitydict = entitydict->next();
  }
}

void Main(Capi *tree)
{ tree->initialise();
  thetree = tree;
  Syntaxtree *tr = tree->syntaxtreeref;
  seterrors(tr->errormessages[6]);
  setentities();
  for (int i = 0;i<(int)REMOTESCOPE;i++)
      setunused(tree->propertiestreeref->selectqualdict((Qualifierelemkind)i));
  tr->traverseallobjects(checkid);
  for (i = 0;i<(int)REMOTESCOPE;i++)
    if (entities[i])
       reportunused(tree->propertiestreeref->selectqualdict((Qualifierelemkind)i));
}

